"""A python module for scalable analysis of Cloud Optimized GeoTIFFs.

COG Worker is a simple library to help you chunk and run large scale analysis
on Cloud Optimized GeoTIFFs (COGS).
"""

__version__ = "0.1.2"

from .worker import Worker  # noqa
from .manager import Manager  # noqa
from .types import *  # noqa



"""Distributed processing with Dask.

The DaskManager class provides an identical interface to the
:obj:`cog_worker.manager.Manager`, but executes tasks in a
`Dask cluster <https://distributed.dask.org/>`_, instead of on your local machine.

Note:
    cog_worker does not include dask.distributed as a dependency by default.
    In order to use cog_worker.distributed you must install dask.distributed::

        pip install dask[distributed]

Example:
    Read a COG in chunks and sum the results::

        from cog_worker.distributed import DaskManager
        from dask.distributed import Client, LocalCluster

        def my_analysis(worker):
            arr = worker.read('example-cog.tif')
            return arr.sum()

        cluster = LocalCluster()
        client = Client(cluster)
        manager = DaskManager(client)

        results = manager.chunk_execute(my_analysis)
        total = sum(results)
"""
from typing import Iterable, Iterator, Mapping, Union, Tuple, Any
import logging

import dask
import dask.distributed
from pyproj import Proj

import cog_worker
from cog_worker.types import WorkerFunction, BoundingBox


logger = logging.getLogger(__name__)


class DaskManager(cog_worker.manager.Manager):
    """Class for chunking and executing cog_worker functions in a dask cluster.

    The DaskManager identical to the cog_worker.manager.Manager, except that it
    executes functions in a Dask cluster instead of locally.
    """

    def __init__(
        self,
        dask_client: dask.distributed.Client,
        bounds: BoundingBox = (-180, -85, 180, 85),
        proj: Union[int, str, Proj] = 3857,
        scale: float = 10000,
        buffer: int = 16,
    ):
        """Initialize a DaskManager with a dask client.

        Args:
            dask_client (dask.distributed.Client): The dask client to use to
                execute analysis.
            bounds (BoundingBox): The region to be analyzed.
            proj (pyproj.Proj, str, int): The projection to analyze in.
                Generally accepts any proj4 string, WKT projection, or EPSG
                code. See pyproj.Proj for valid values.
            scale (float): The pixel size for analysis in the projection's units
                (usually meters or degrees).
            buffer (int): When dividing analysis into chunks, the number of
                additional pixels to read on all sides to avoid edge effects.
                The ideal buffer size depends on your analysis (e.g. whether you
                use convolutions or distance functions).
        """
        self.client = dask_client
        super().__init__(bounds, proj, scale, buffer)

    def execute(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        clip: bool = True,
        **kwargs
    ) -> Tuple[Any, BoundingBox]:
        """Execute a cog_worker function in the DaskManager's cluster.

        The execute method is the underlying method for running analysis. By
        default, it will run the function over the Manager's bounding box in a
        single chunk.

        When executing functions, the Manager instantiates a
        cog_worker.worker.Worker and passes it to the function as its first
        parameter. The Worker keeps track of the scale, projection, and bounds
        of its piece of the analysis, which it uses to handle the reading and
        writing of Cloud Optimized GeoTIFFs.

        Args:
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will
                recieve a cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the
                function.
            clip (bool): Whether or not to clip the `buffer` from the completed
                analysis.
            **kwargs: Additional keyword arguments to overload the Manager's
                properties. (bounds, proj, scale, or buffer)

        Returns:
            A tuple containing the return value of the function and the bounding
            box of the executed analysis in the target projection.
        """
        args = {
            "f": f,
            "f_args": f_args,
            "f_kwargs": f_kwargs,
            "bounds": self.bounds,
            "proj": self.proj,
            "scale": self.scale,
            "buffer": self.buffer,
            "clip": clip,
        }
        args.update(kwargs)
        future = self.client.submit(cog_worker.manager._execute, **args)
        return future.result()

    def chunk_execute(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        chunksize: int = 512,
    ) -> Iterator[Tuple[Any, BoundingBox]]:
        """Compute chunks in parallel in the DaskManager's cluster.

        Chunks will be yielded as they are completed. The order in which they
        are yielded is not guaranteed.

        Note:
            You can estimate the memory requirement of executing a function at a
            given chunksize as:
            ``(chunksize + 2*buffer)**2 * number_of_bands_or_arrays * bit_depth``.

        Args:
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will
                recieve a cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the
                function.
            chunksize (int): Size of the chunks in pixels (excluding buffer).

        Yields:
            A tuple containing the return value of the function for each chunk
            and the bounding box of the executed analysis in the target
            projection.
        """
        tasks = [
            dask.delayed(cog_worker.manager._execute)(f, f_args, f_kwargs, **params)
            for params in self.chunk_params(chunksize)
        ]
        futures = self.client.compute(tasks)

        for future, result in dask.distributed.as_completed(futures, with_results=True):
            yield result



"""Previewing, chunking, and executing analysis.

The Manager class is used to divide an area of analysis into chunks of manageable size,
and execute functions on each of these chunks.

When executing functions, the Manager instantiates a :obj:`cog_worker.worker.Worker` and passes
it to the function as its first parameter. The Worker keeps track of the scale, projection,
and bounds of its piece of the analysis, which it uses to handle the reading and writing of
Cloud Optimized GeoTIFFs.

Example:
    Use the manager to preview an analysis before executing it::

        from cog_worker import Manager
        from rasterio.plot import show
        def my_analysis(worker):
            arr = worker.read('example-cog.tif')
            # calculations ...
            return arr

        manager = Manager()
        arr, bbox = manager.preview(my_analysis)
        show(arr)

    Execute the analysis in chunks, saving the results to disk::

        manager.chuck_save('output.tif', myanalysis):

"""
import math
import logging
from typing import IO, Iterable, Iterator, Mapping, Optional, Tuple, Type, Union, Any


import numpy as np
import morecantile
from pyproj import Proj
from rasterio.crs import CRS
import rasterio as rio
from rasterio.io import DatasetWriter
import rasterio.windows

import cog_worker.worker
from .utils import _bbox_size, _get_profile
from .types import WorkerFunction, BoundingBox

logger = logging.getLogger(__name__)


class Manager:
    """Class for managing scalable analysis of Cloud Optimized GeoTIFFs."""

    def __init__(
        self,
        bounds: BoundingBox = (-180, -85, 180, 85),
        proj: Union[int, str, Proj] = 3857,
        scale: float = 10000,
        buffer: int = 16,
    ):
        """Initialize a Manager with a projection, scale, and bounding box for analysis.

        Args:
            bounds (BoundingBox): The region to be analyzed as a (west, south,
                east, north) tuple.
            proj (pyproj.Proj, str, int): The projection to analyze in.
                Generally accepts any proj4 string, WKT projection, or EPSG
                code. See pyproj.Proj for valid values.
            scale (float): The pixel size for analysis in the projection's units
                (usually meters or degrees).
            buffer (int): When dividing analysis into chunks, the number of additional pixels
                to read on all sides to avoid edge effects. The ideal buffer size depends on
                your analysis (e.g. whether you use convolutions or distance functions).
        """
        self.proj = proj if isinstance(proj, Proj) else Proj(proj, preserve_units=False)
        self.bounds = bounds
        self.scale = scale
        self.buffer = buffer
        self._proj_bounds = self.proj.transform_bounds(*bounds)
        self.tms = morecantile.TileMatrixSet.custom(
            list(self._proj_bounds), CRS.from_user_input(proj)
        )

    def execute(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        clip: bool = True,
        **kwargs
    ) -> Tuple[Any, BoundingBox]:
        """Execute a function that takes a cog_worker.worker.Worker as its first parameter.

        The execute method is the underlying method for running analysis. By default, it
        will run the function once for the Manager's given scale and bounding box.

        When executing functions, the Manager instantiates a cog_worker.worker.Worker and passes
        it to the function as its first parameter. The Worker keeps track of the scale, projection,
        and bounds of its piece of the analysis, which it uses to handle the reading and writing of
        Cloud Optimized GeoTIFFs.

        Args:
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will recieve a
                cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the function.
            clip (bool): Whether or not to clip the buffer from the completed analysis.
            **kwargs: Additional keyword arguments to overload the Manager's properties.
                (bounds, proj, scale, or buffer)

        Returns:
            A tuple containing the return value of the function and the bounding
            box of the executed analysis in the target projection.
        """
        args = {
            "bounds": self.bounds,
            "proj": self.proj,
            "scale": self.scale,
            "buffer": self.buffer,
        }
        args.update(kwargs)
        return _execute(f, f_args, f_kwargs, clip, **args)

    def preview(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        bounds: Optional[BoundingBox] = None,
        max_size: int = 1024,
        **kwargs
    ) -> Tuple[Any, BoundingBox]:
        """Preview a function by executing it at a reduced scale.

        The preview method automatically reduces the scale of analysis to fit within `max_size`.

        Args:
            f (WorkerFunction): The function to execute. The function will
                recieve a cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the function.
            bounds (BoundingBox, default: self.bounds): The region to analize.
            max_size (int): The maximum size (width or height) in pixels to
                compute, ignoring any buffer (default: 1024px).
            **kwargs: Additional keyword arguments to overload the Manager's properties.
                (proj or buffer).

        Returns:
            A tuple containing the return value of the function and the bounding
            box of the executed analysis in the target projection.
        """
        bounds = self.bounds if bounds is None else bounds
        proj = kwargs.pop("proj", self.proj)
        proj = proj if isinstance(proj, Proj) else Proj(proj, preserve_units=False)
        proj_bounds = self.proj.transform_bounds(*bounds)
        width, height = _bbox_size(proj_bounds, self.scale)
        _size = max(width, height)
        scale = self.scale * _size / max_size

        kwargs.update({"proj_bounds": proj_bounds, "proj": proj, "scale": scale})

        return self.execute(f, f_args, f_kwargs, **kwargs)

    def tile(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        z: int = 0,
        x: int = 0,
        y: int = 0,
        tilesize: int = 256,
        **kwargs
    ) -> Tuple[Any, BoundingBox]:
        """Execute a function for the scale and bounds of a TMS tile.

        The tile method supports non-global and non-mercator tiling schemes via
        Morecantile. To generate standard web tiles, instantiate the Manager
        with the default parameters.

        Args:
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will
                recieve a cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the function.
            bounds (BoundingBox): The region to analize (default: self.bounds)
            max_size (int): The maximum size (width or height) in pixels to compute, ignoring any buffer
                (default: 1024px). Automatically reduces the scale of analysis to fit within `max_size`.
            **kwargs: Additional keyword arguments to overload the Manager's properties.
                (buffer).

        Returns:
            A tuple containing the return value of the function and the bounding
            box of the executed analysis in the target projection.
        """
        proj_bounds = self.tms.xy_bounds(x, y, z)  # type: ignore
        left, bottom, right, top = proj_bounds
        size = max(right - left, top - bottom)
        scale = size / tilesize

        kwargs.update(
            {
                "proj_bounds": proj_bounds,
                "scale": scale,
            }
        )

        return self.execute(f, f_args, f_kwargs, **kwargs)

    def chunk_execute(
        self,
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        chunksize: int = 512,
    ) -> Iterator[Tuple[Any, BoundingBox]]:
        """Return a generator that executes a function on chunks of at most `chunksize` pixels.

        Note:
            Manager.chunk_execute computes each chunk sequentially, trading time for reduced memory footprint.
            To run large scale analysis in parallel using dask, see cog_worker.distributed.

        Note:
            You can estimate the memory requirement of executing a function at a given chunksize as
            ``(chunksize + 2*buffer)**2 * number_of_bands_or_arrays * bit_depth``.

        Args:
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will recieve a
                cog_worker.worker.Worker as its first argument.
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the function.
            chunksize (int): Size of the chunks in pixels (excluding buffer).

        Yields:
            A tuple containing the return value of the function and the bounding
            box of the executed analysis in the target projection.
        """
        for params in self.chunk_params(chunksize):
            yield self.execute(f, f_args, f_kwargs, **params)

    def chunk_save(
        self,
        dst: Union[str, IO],
        f: WorkerFunction,
        f_args: Iterable = None,
        f_kwargs: Mapping = None,
        chunksize: int = 512,
        **kwargs
    ):
        """Execute a function in chunks and write each chunk to disk as it is completed.

        The chunk_save method is identical to Manager.chunk_execute, except it writes results to ``dst``
        instead of yielding them. Manager.chunk_save uses the rasterio GeoTiff driver.

        Note:
            The function to be executed will recieve a cog_worker.worker.Worker as its first argument and
            should return a 3-dimensional numpy array of ``chunksize`` (optionally plus the buffer pixels).
            e.g.::

                # Read a cog in chunks and write those chunks to 'test.tif'
                manager.chunk_save('test.tif', lambda worker: worker.read('example-cog-url.tif'))

        Args:
            dst (str): The file path to write to.
            f (:obj:`cog_worker.types.WorkerFunction`): The function to execute.
                The function will recieve a cog_worker.worker.Worker as its first argument
                and must return a 3-dimensional numpy array of ``chunksize`` (including or excluding the buffer).
            f_args (list): Additional arguments to pass to the function.
            f_kwargs (dict): Additional keyword arguments to pass to the function.
            chunksize (int): Size of the chunks in pixels (excluding buffer).
            **kwargs: Additional keyword arguments to pass to rasterio.open.
        """
        chunks = self.chunk_execute(f, f_args, f_kwargs, chunksize)
        arr, bbox = next(chunks)
        with self._open_writer(dst, arr.shape[0], arr.dtype, **kwargs) as _writer:
            self._write_chunk(_writer, arr, bbox)
            for arr, bbox in chunks:
                self._write_chunk(_writer, arr, bbox)

    def _open_writer(
        self, dst: Union[str, IO], count: int, dtype: Type, **kwargs
    ) -> DatasetWriter:
        """Open a rasterio.DatasetWriter with default profile."""
        profile = _get_profile(
            count, self.scale, self._proj_bounds, self.proj, dtype, **kwargs
        )

        return rio.open(dst, "w", **profile)  # type: ignore

    def _write_chunk(
        self,
        writer: DatasetWriter,
        arr: np.ndarray,
        bbox: BoundingBox,
    ):
        """Write a chunk to a rasterio.DatasetWriter."""
        if len(arr.shape) == 2:
            arr = arr[np.newaxis]
        height, width = arr.shape[1:]
        window = rasterio.windows.from_bounds(*bbox, writer.transform, height, width)

        writer.write(arr, window=window)
        if isinstance(arr, np.ma.MaskedArray):
            mask = np.ma.getmask(arr)
            if len(arr.shape) == 3:
                mask = np.any(mask, axis=0)
            writer.write_mask(mask, window=window)

    def chunk_params(self, chunksize: int = 512, **kwargs):
        """Generate parameters to execute a function in chunks.

        Generates dicts of keyword arguments that can be passed to Manager.execute to run a function in chunks
        of size <chunksize>. This may be useful for distributing tasks to workers to execute in parallel. Each dict
        will contain the projection, scale, bounding box, and buffer. Attributes will be identical except
        for ``proj_bounds`` which define the area to analyze.

        Note:
            ``manager.chunk_execute(f)`` is equivalent to
            ``(manager.execute(f, **params) for params in manager.chunk_params())``

        Args:
            chunksize (int): Size of the chunks in pixels (excluding buffer).
            **kwargs: optional additional keyword arguments to save to the dict (to eventually pass to Manager.execute)
                e.g. ``f``, ``f_args``, ``f_kwargs``

        Yields:
            Dicts of keyword arguments that can be passed to :obj:`cog_worker.manager.Manager.execute()`.
        """
        _args = {
            "proj": self.proj.srs,
            "scale": self.scale,
            "buffer": self.buffer,
        }
        _args.update(kwargs)

        for proj_bounds in self.chunks(chunksize):
            args = _args.copy()
            args["proj_bounds"] = proj_bounds
            yield args

    def chunks(self, chunksize: int = 512) -> Iterator[BoundingBox]:
        """Generate bounding boxes for chunks of at most <chunksize> pixels in the managers scale and projection.

        The chunks method divides the Manager's bounding box into chunks of manageable size.
        Each chunk will be at most <chunksize> pixels, though the geographic extent of the chunk
        depends on the Manager's projection and scale.

        Args:
            chunksize (int): Size of the chunks in pixels (excluding buffer).

        Yields:
            BoundingBox: The bounding box of the chunk in the Manager's projection
        """
        xshards, yshards = self._num_chunks(chunksize)
        for i in range(xshards):
            for j in range(yshards):
                bounds = self._chunk_bounds(i, j, chunksize)
                if np.isfinite(bounds).all():
                    yield bounds

    def _chunk_bounds(
        self,
        x: int,
        y: int,
        chunksize: int,
    ) -> BoundingBox:
        """Get the bounding box of a chunk with index <x>,<y>."""
        left, bottom, right, top = self._proj_bounds
        _chunksize = chunksize * self.scale

        l = left + x * _chunksize
        r = min(l + _chunksize, right)
        t = top - y * _chunksize
        b = max(t - _chunksize, bottom)

        return (l, b, r, t)

    def _num_chunks(
        self,
        chunksize: int,
    ) -> Tuple[int, int]:
        """Return the number of chunks necessary to cover the Manager's bounding box."""
        left, bottom, right, top = self._proj_bounds
        return (
            math.ceil((right - left) / self.scale / chunksize),
            math.ceil((top - bottom) / self.scale / chunksize),
        )


def _execute(
    f: WorkerFunction,
    f_args: Iterable = None,
    f_kwargs: Mapping = None,
    clip: bool = True,
    **kwargs
) -> Tuple[Any, BoundingBox]:
    """Execute a function that takes a cog_worker.worker.Worker as its first parameter.

    Instantiate a cog_worker.worker.Worker and pass it to the function as its first parameter.

    Args:
        f (:obj:`cog_worker.types.WorkerFunction`): The function to execute. The function will recieve a
            cog_worker.worker.Worker as its first argument.
        f_args (list): Additional arguments to pass to the function.
        f_kwargs (dict): Additional keyword arguments to pass to the function.
        clip (bool): Whether or not to clip the buffer from the completed analysis.
        **kwargs: Additional keyword arguments to instantiate the cog_worker.worker.Worker

    Returns:
        A tuple containing the return value of the function and the bounding
        box of the executed analysis in the target projection.
    """
    worker = cog_worker.worker.Worker(**kwargs)

    f_args = [] if f_args is None else f_args
    f_kwargs = {} if f_kwargs is None else f_kwargs

    arr: np.ndarray = f(worker, *f_args, **f_kwargs)  # type: ignore
    if clip and isinstance(arr, np.ndarray):
        arr = worker.clip_buffer(arr)

    return arr, worker.bounds



"""Utility functions."""
from typing import Tuple, Type, Union

from rasterio import transform
from pyproj import Proj
import numpy as np

from .types import BoundingBox


def _get_profile(
    count: int,
    scale: float,
    proj_bounds: BoundingBox,
    proj: Proj,
    dtype: Union[Type, np.dtype],
    **kwargs
) -> dict:
    width, height = _bbox_size(proj_bounds, scale)
    affine = transform.from_origin(proj_bounds[0], proj_bounds[3], scale, scale)
    profile = {
        "driver": "GTiff",
        "interleave": "pixel",
        "blockxsize": 512,
        "blockysize": 512,
        "tiled": True,
        "compress": "lzw",
        "crs": proj.srs,
        "transform": affine,
        "dtype": dtype,
        "width": width,
        "height": height,
        "count": count,
    }
    profile.update(kwargs)
    return profile


def _bbox_size(
    bounds: BoundingBox,
    scale: float,
) -> Tuple[int, int]:
    left, bottom, right, top = bounds
    return (round((right - left) / scale), round((top - bottom) / scale))



"""cog_worker type definitions."""
from typing import Callable, Tuple, Union

import numpy as np

import cog_worker.worker

BoundingBox = Tuple[float, float, float, float]
"""A ``(west, south, east, north)`` tuple."""

WorkerFunction = Union[
    Callable[["cog_worker.worker.Worker"], np.ndarray],
    Callable,
]
"""A function that can recieve a cog_worker.worker.Worker as its first parameter.

Additional aguments and keyword arguments can be passed to the Worker function
at time of execution with the ``f_args`` and ``f_kwargs`` parameters of
:obj:`cog_worker.manager.Manager.execute()`

Example:

    Read a specific COG and return it as an array::

        def my_analysis(worker: cog_worker.Worker):
            arr = worker.read('example-cog.tif')
            return arr

    Read a COG at a given url and get the neighborhood mean for a 1km square kernel::

        from scipy.ndimage import uniform_filter

        def my_analysis(worker: cog_worker.Worker, source_url: str):
            arr = worker.read(source_url)
            kernel_size = 1000/worker.scale  # in map units (meters)
            return uniform_filter(arr, kernel_size)

    Read a COG and optionally upload each chunk to an S3 bucket as it is computed::

        from rasterio import MemoryFile
        import boto3

        def my_analysis(worker: cog_worker.Worker, dst_bucket: str):
            arr = worker.read('example-cog.tif')

            if dst_bucket:
                with MemoryFile() as memfile:
                    fname = f'output_{worker.scale}_{worker.bounds[0]}_{worker.bounds[3]}.tif'
                    worker.write(arr, memfile)

                    memfile.seek(0)
                    boto3.client('s3').upload_fileobj(memfile, dst_bucket, fname)

            return arr
"""



#!/usr/bin/env python
from setuptools import setup

with open("README.md") as f:
    desc = f.read()

setup(
    name="cog_worker",
    version="0.1.2",
    description="Scalable geospatial analysis on Cloud Optimized GeoTIFFs.",
    long_description=desc,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Francis Gassert",
    author_email="francis.gassert@vizzuality.com",
    url="https://github.com/vizzuality/cog_worker",
    packages=["cog_worker"],
    keywords="cog geotiff raster gdal rasterio dask",
    install_requires=[
        "numpy",
        "pyproj",
        "rasterio",
        "morecantile",
        "rio_tiler",
    ],
    extras_require={
        "test": ["pytest"],
        "dev": ["pre-commit"],
        "distributed": ["dask[distributed]"],
        "docs": [
            "Sphinx",
            "sphinxcontrib-napoleon",
            "furo",
            "nbsphinx",
            "nbconvert"
        ],
    },
)



"""Reading COGs.

The Worker class keeps track of the region, projection, and scale to
conduct analysis in.

When writing cog_worker functions, the main method you will use is
:obj:`Worker.read()`, which is a wrapper around ``rio_tiler`` to clip,
reproject and resample the data into the target resolution.

Example:
    Read a COG, reprojecting it onto a global 1-deg lat-long grid::

        from cog_worker import Worker
        from rasterio.plot import show

        worker = Worker(bounds=(-180, -90, 180, 90), proj=4326, scale=1.0)
        arr = worker.read('example-cog-url.tif')
        show(arr)
"""
import logging
from typing import Sequence, Union

from pyproj import Proj
from pyproj.enums import TransformDirection
import rasterio as rio
from rasterio._err import CPLE_AppDefinedError
import numpy as np
from rio_tiler.errors import EmptyMosaicError
from rio_tiler.models import ImageData
from rio_tiler.io import COGReader
from rio_tiler.mosaic import mosaic_reader

from cog_worker.types import BoundingBox
from cog_worker.utils import _bbox_size, _get_profile

logger = logging.getLogger(__name__)


class Worker:
    """Class for reading Cloud Optimized GeoTIFFs."""

    def __init__(
        self,
        bounds: BoundingBox = (-180, -85, 180, 85),
        proj_bounds: BoundingBox = None,
        proj: Union[int, str, Proj] = 3857,
        scale: float = 10000,
        buffer: int = 16,
    ):
        """Initialize a Worker with a bounding box, scale, and projection.

        Args:
            bounds (BoundingBox): The region to be analyzed as a (west, south, east, north) tuple.
                Ignored when ``proj_bounds`` is provided.
            proj_bounds (BoundingBox): The region to be analyzed in the Worker's projection.
                Overrides ``bounds`` when provided.
            proj (pyproj.Proj, str, int): The projection to analyze in. See
                ``pyproj.Proj`` for valid values (https://pyproj4.github.io/pyproj/).
            scale (float): The pixel size for analysis in the projection's units (usually meters or degrees).
            buffer (int): The number of additional pixels to read on all sides to avoid edge effects.
                The ideal buffer size depends on your analysis (e.g. whether you plan to use convolutions or
                distance functions).
        """
        self._proj = (
            proj if isinstance(proj, Proj) else Proj(proj, preserve_units=False)
        )

        if proj_bounds is None:
            proj_bounds = self._proj.transform_bounds(*bounds)

        self._bounds = proj_bounds
        self._scale = scale
        self._buffer = buffer

        self._width, self._height = _bbox_size(self._bounds, scale)

    @property
    def proj(self) -> Proj:
        """The projection used for reading."""
        return self._proj

    @property
    def bounds(self) -> BoundingBox:
        """The the bounding box in projected coordinates."""
        return self._bounds

    @property
    def scale(self) -> float:
        """The size of pixels in projection units."""
        return self._scale

    @property
    def buffer(self) -> int:
        """The number of additional pixels to read on all sides of the Worker's bounding box."""
        return self._buffer

    @property
    def width(self) -> int:
        """The width of the Worker's bounding box in pixels."""
        return self._width

    @property
    def height(self) -> int:
        """The height of the Worker's bounding box in pixels."""
        return self._height

    def xy_bounds(self, buffered: bool = False) -> BoundingBox:
        """Return the Worker's bounding box in projected coordinates.

        Args:
            buffered (bool): Buffer the worker's bounding box
        """
        return self._buffer_bbox() if buffered else self.bounds

    def lnglat_bounds(self, buffered: bool = False) -> BoundingBox:
        """Return the Worker's bounding box in geographic coordinates.

        Note:
            When using a projected coordinate system, the geographic bounding box
            that covers the Worker's projected extent may be larger
            than the bounding box used to instantiate the Worker.

        Args:
            buffered (bool): Buffer the Worker's bounding box.
        """
        pts = max(self.width, self.height) + (buffered * self.buffer * 2) - 1
        bounds = self.xy_bounds(buffered)
        return self.proj.transform_bounds(
            *bounds, pts, direction=TransformDirection.INVERSE
        )

    def empty(self, mask: bool = False) -> np.ndarray:
        """Return a zeroed array covering the Worker's extent including the buffer.

        Args:
            mask (bool): Return a Numpy masked array with all pixels masked.
                Otherwise returns a standard Numpy array filled with zeros.
        """
        arr = np.zeros((1, self.height + self.buffer * 2, self.width + self.buffer * 2))
        if mask:
            _mask = np.ones(
                (self.height + self.buffer * 2, self.width + self.buffer * 2)
            )
            arr = np.ma.array(arr, mask=_mask)
        return arr

    def read(self, src: Union[str, Sequence[str]], **kwargs) -> np.ma.MaskedArray:
        """Read a COG, reprojecting and clipping as necessary.

        The read method uses ``rio_tiler.COGReader`` to takes advantage of the
        file structure and internal overviews in COGs, minimizing the amount of
        data that needs to be read and transferred when working at reduced resolutions.

        In general, any valid GDAL path can be read. This may be a url pointing to a COG, a local
        GeoTIFF or a GDAL virtual file system path. However, it may be very inefficient to
        read data sources that are not valid Cloud Optimized GeoTIFFs.

        If a list of data sources is provided, ``Worker.read`` will use ``rio_tiler.mosaic_reader``
        to mosaic the sources together.

        Note:
            The resampling method used to generate the COG's internal overviews will affect
            how it appears at reduced resolutions.

        Args:
            src (str, list): The data source to read or list of sources to mosiac.
            **kwargs: Additional keyword arguments to pass to ``rio_tiler.COGReader.part``
                or ``rio_tiler.mosaic_reader``. See: https://cogeotiff.github.io/rio-tiler/.

        Returns:
            A Numpy masked array containing the data for the Worker's bounding box and its
            buffer.

        Note:
            The mask values of a Numpy masked array is the inverse of a GDAL (alpha) mask.
            A masked value of True corresponds to nodata or an alpha value of 0.
        """
        proj_bounds = self._buffer_bbox()
        width, height = _bbox_size(proj_bounds, self._scale)

        if isinstance(src, str):
            img = _read_COG(src, proj_bounds, self._proj.srs, width, height, **kwargs)
        elif isinstance(src, Sequence):
            try:
                img, asset = mosaic_reader(
                    src, _read_COG, proj_bounds, self.proj.srs, width, height, **kwargs
                )
            except EmptyMosaicError:
                return self.empty(mask=True)

        arr = img.data
        mask = (img.mask == 0) | np.isnan(arr)

        return np.ma.array(arr, mask=mask)

    def write(self, arr: np.ndarray, dst: str, **kwargs):
        """Write a Numpy array to a GeoTIFF.

        The write method will create a GeoTIFF with a profile matching the Worker's properties.
        Uses ``rasterio.open`` under the hood.

        Args:
            arr (numpy.ndarray): The array to write. Must be 2 or 3-dimensional, with a width and
                height matching the Worker (including or excluding the buffer). If the array
                includes the Worker's buffer, the buffer will be clipped before writing.
            dst (str): The file path to write to.
            **kwargs: Additional keyword arguments to pass to rasterio.open
                See: https://rasterio.readthedocs.io/en/latest/topics/writing.html
        """
        arr = self.clip_buffer(arr)
        count, height, width = arr.shape
        profile = _get_profile(
            count, self.scale, self._bounds, self.proj, arr.dtype, **kwargs
        )

        with rio.open(dst, "w", **profile) as writer:
            writer.write(arr)
            if isinstance(arr, np.ma.MaskedArray):
                mask = np.ma.getmask(arr)
                if len(arr.shape) == 3:
                    mask = np.any(mask, axis=0)
                writer.write_mask(mask)

    def clip_buffer(self, arr: np.ndarray) -> np.ndarray:
        """Clip the buffer pixels from an array if they exist.

        Args:
            arr (numpy.ndarray): The array to clip.

        Returns:
            The array with buffer pixels removed.

        Raises:
            ValueError: If the array's shape does not match the Worker's width and height
        """
        if len(arr.shape) == 2:
            arr = arr[np.newaxis]
        buffer_width = self.width + self.buffer * 2
        buffer_height = self.height + self.buffer * 2
        c, h, w = arr.shape
        if w == self.width and h == self.height:
            return arr
        elif w == buffer_width and h == buffer_height:
            return arr[:, self.buffer : -self.buffer, self.buffer : -self.buffer]
        else:
            raise ValueError(
                f"Array not expected size. Was {w}x{h} expected {self.width}x{self.height} or {buffer_width}x{buffer_height}"
            )

    def _buffer_bbox(self) -> BoundingBox:
        """Returns the worker's bounding box extended by the buffered pixels."""
        l, b, r, t = self._bounds
        _buffer = self.buffer * self.scale

        return (l - _buffer, b - _buffer, r + _buffer, t + _buffer)


def _read_COG(
    asset: str,
    proj_bounds: BoundingBox,
    crs: Union[str, int, Proj],
    width: int,
    height: int,
    **kwargs,
) -> ImageData:
    """Read part of a COG, warping and resampling to a target shape."""
    tries = 0
    while tries < 5:
        try:
            with COGReader(filepath=asset, **kwargs) as cog:  # type: ignore
                return cog.part(
                    proj_bounds,
                    bounds_crs=crs,
                    dst_crs=crs,
                    max_size=None,
                    width=width,
                    height=height,
                )
        except CPLE_AppDefinedError as e:
            # Ignore some strange GDAL errors when reading in some projections
            # see: https://rasterio.groups.io/g/main/message/780
            logger.debug(e)
    raise Exception(f"Failed reading asset {asset}")
